/* Undefined Instruction handling
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Image Loader.

The Imote2 Image Loader is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

The Imote2 Image Loader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the Imote2 Image Loader. If not, see <http://www.gnu.org/licenses/>.
*/
#include "exception.h"
#include "exceptions.h"
#include "ledcycle.h"
#include "doublefault.h"
#include "backtrace.h"
#include <stdio.h>

static void printFault(uint32_t lr, const uint32_t * fp,
	const struct _ExcProcessorStatus * pstat);

__attribute__((interrupt("UNDEF"),weak))
void hplarmv_undef() {
	/* gather some data: lr, fp, fault source, fault address and processor status */
	uint32_t lr = ((uint32_t)__builtin_return_address(0));
	uint32_t * fp = (uint32_t*)__builtin_frame_address(0);
	struct _ExcProcessorStatus pstat;
	_excGetProcessorStatus(&pstat);
	
	/* check for double faults */
	_excDoubleFaultCheck(lr, fp, &pstat);
	
	/* initialize exception handler */
	_excInit();
	
	/* begin in next line */
	_excUartOut("\n");
	
	while (1) {
		printFault(lr, fp, &pstat);
		_excLedCycle(TRUE);
	}
}

static void printFault(uint32_t lr, const uint32_t * fp,
	const struct _ExcProcessorStatus * pstat) {
	char buf[128];
	
	/* print the exception type and the old mode */
	snprintf(buf, sizeof(buf), "UNDEFINED INSTRUCTION occured in %s mode\n", pstat->spsrModeLabel);
	_excUartOut(buf);
	
	snprintf(buf, sizeof(buf), " Failed OpCode was %08x at 0x%08x\n", *(uint32_t*)(lr - 4), lr - 4);
	_excUartOut(buf);
	
	_excPrintProcessorStatus(lr, fp, pstat);
	
	_excBacktrace(lr, fp);
}

